from django.conf import settings
from django.contrib import admin
from jauth.models import (
    AccountKitUser,
    AccountKitAccessToken,
    FacebookAccessToken,
    FacebookUser,
    GoogleUser,
    GoogleAccessToken,
)


class JauthAdminBase(admin.ModelAdmin):
    pass


class AccountKitAdmin(JauthAdminBase):
    pass


class AccountKitUserAdmin(AccountKitAdmin):
    raw_id_fields = [
        "user",
    ]


class FacebookAdmin(JauthAdminBase):
    pass


class GoogleAdmin(JauthAdminBase):
    pass


class FacebookUserAdmin(FacebookAdmin):
    raw_id_fields = [
        "user",
    ]


class GoogleUserAdmin(GoogleAdmin):
    raw_id_fields = [
        "user",
    ]


if hasattr(settings, "ACCOUNT_KIT_APP_ID") and settings.ACCOUNT_KIT_APP_ID:
    admin.site.register(AccountKitUser, AccountKitUserAdmin)
    admin.site.register(AccountKitAccessToken, AccountKitAdmin)
if hasattr(settings, "FACEBOOK_APP_ID") and settings.FACEBOOK_APP_ID:
    admin.site.register(FacebookUser, FacebookUserAdmin)
    admin.site.register(FacebookAccessToken, FacebookAdmin)
if hasattr(settings, "GOOGLE_APP_ID") and settings.GOOGLE_APP_ID:
    admin.site.register(GoogleUser, GoogleUserAdmin)
    admin.site.register(GoogleAccessToken, GoogleAdmin)

required_params = ["JAUTH_AUTHENTICATION_ERROR_REDIRECT", "JAUTH_AUTHENTICATION_SUCCESS_REDIRECT"]
for p in required_params:
    if not hasattr(settings, p):
        raise Exception("{} configuration missing".format(p))



from django.contrib.auth.models import User
from django.test import TestCase

from jauth.serializers import AuthResultSerializer


class Tests(TestCase):
    def add_test_user(self, email: str = "test@example.com", password: str = "test1234"):
        self.user = User.objects.create_user(email, email, password)

    def setUp(self):
        self.add_test_user()

    def tearDown(self):
        pass

    def test_serializers(self):
        data = AuthResultSerializer(self.user).data
        self.assertIn("token", data)
        self.assertIn("user", data)
        self.assertIn("id", data["user"])
        self.assertIn("username", data["user"])
        self.assertIn("email", data["user"])
        self.assertDictEqual(data["user"], {"id": self.user.id, "email": self.user.email, "username": self.user.username})



# pylint: disable=logging-format-interpolation
import base64
import hashlib
import hmac
import json
import logging
from urllib.parse import urlencode

import requests
from django.conf import settings
from requests import Response
from rest_framework.exceptions import ValidationError


logger = logging.getLogger(__name__)


def http_request(method: str, url: str, **kwargs) -> Response:
    res = requests.request(method, url, **kwargs)
    logger.info("{} {} -------------------------------------------------------------------- ".format(method.upper(), url))
    logger.info("{}".format(res.request.body))  # type: ignore
    logger.info("HTTP {} -------------------------------------------------------------------- ".format(res.status_code))
    logger.info("{}".format(res.text))
    return res


def account_kit_get_access_token(code: str) -> Response:
    """
    GET https://graph.accountkit.com/v1.3/access_token?grant_type=authorization_code&code=<authorization_code>&access_token=AA|<facebook_app_id>|<app_secret>
    :param code:
    :return: dict
    """
    app_id = settings.ACCOUNT_KIT_APP_ID
    app_secret = settings.ACCOUNT_KIT_APP_SECRET
    url = settings.ACCOUNT_KIT_API_URL + "/access_token?grant_type=authorization_code&code={}&access_token=AA|{}|{}".format(code, app_id, app_secret)
    res = http_request("get", url)
    return res


def account_kit_me(access_token: str) -> Response:
    """
    GET https://graph.accountkit.com/v1.3/me/?access_token=<access_token>
    :param access_token:
    :return: str
    """
    url = settings.ACCOUNT_KIT_API_URL + "/me?access_token={}".format(access_token)
    res = http_request("get", url)
    return res


def facebook_get_access_token(code: str) -> Response:
    """
    :param code:
    :return: dict
    """
    app_id = settings.FACEBOOK_APP_ID
    app_secret = settings.FACEBOOK_APP_SECRET
    redirect_url = settings.FACEBOOK_REDIRECT_URL
    url = settings.FACEBOOK_API_URL + "/oauth/access_token?client_id={}&redirect_uri={}&client_secret={}&code={}".format(
        app_id, redirect_url, app_secret, code
    )
    res = http_request("get", url)
    return res


def facebook_me(access_token: str) -> Response:
    """
    :param access_token:
    :return: str
    """
    url = settings.FACEBOOK_API_URL + "/me?access_token={}".format(access_token)
    res = http_request("get", url)
    return res


def facebook_parse_signed_request(signed_request: str) -> dict:
    encoded_sig, payload = signed_request.split(".", 2)
    app_secret = settings.FACEBOOK_APP_SECRET
    data = json.loads(base64.b64decode(payload.replace("-_", "+/")))
    expected_sig = hmac.new(app_secret.encode(), msg=payload.encode(), digestmod=hashlib.sha256).digest()
    if expected_sig != encoded_sig:
        raise ValidationError("Expected signature does not match encoded signature in Facebook request")
    return data


def google_get_access_token(code: str) -> Response:
    """
    :param code:
    :return: dict
    """
    app_id = settings.GOOGLE_APP_ID
    app_secret = settings.GOOGLE_APP_SECRET
    redirect_url = settings.GOOGLE_REDIRECT_URL
    url = settings.GOOGLE_API_URL + "/oauth2/v4/token"
    data = {
        "code": code,
        "client_id": app_id,
        "client_secret": app_secret,
        "redirect_uri": redirect_url,
        "grant_type": "authorization_code",
    }
    res = http_request("post", url, data=data)
    return res


def google_me(access_token: str) -> Response:
    """
    {
      "id": "107873301371187782400",
      "email": "kajala@gmail.com",
      "verified_email": true,
      "name": "Jani Kajala",
      "given_name": "Jani",
      "family_name": "Kajala",
      "picture": "https://lh4.googleusercontent.com/-g1HKAdB0pi0/AAAAAAAAAAI/AAAAAAACcbQ/pCyIjkFBHv8/photo.jpg",
      "locale": "en"
    }
    :param access_token:
    :return: str
    """
    url = settings.GOOGLE_API_URL + "/oauth2/v1/userinfo?alt=json&access_token={}".format(access_token)
    res = http_request("get", url)
    return res


def google_login_url(state: str, scope: str = "email profile", api_url: str = "https://accounts.google.com/o/oauth2/v2/auth", **kw) -> str:
    return (
        api_url
        + "?"
        + urlencode(
            {
                "client_id": settings.GOOGLE_APP_ID,
                "redirect_uri": settings.GOOGLE_REDIRECT_URL,
                "state": state,
                "scope": scope,
                "response_type": "code",
                **kw,
            }
        )
    )


def facebook_login_url(state: str, api_url: str = "https://www.facebook.com/v3.3/dialog/oauth", **kw) -> str:
    return (
        api_url
        + "?"
        + urlencode(
            {
                "client_id": settings.FACEBOOK_APP_ID,
                "redirect_uri": settings.FACEBOOK_REDIRECT_URL,
                "state": state,
                **kw,
            }
        )
    )


def account_kit_sms_login_url(state: str, api_url: str = "https://www.accountkit.com/v1.0/basic/dialog/sms_login/", **kw) -> str:
    return (
        api_url
        + "?"
        + urlencode(
            {
                "app_id": settings.ACCOUNT_KIT_APP_ID,
                "redirect": settings.ACCOUNT_KIT_REDIRECT_URL,
                "state": state,
                "fbAppEventsEnabled": "true",
                **kw,
            }
        )
    )


def oauth2_login_urls(state: str, **kw) -> dict:
    data = {}
    if hasattr(settings, "ACCOUNT_KIT_APP_ID") and settings.ACCOUNT_KIT_APP_ID:
        data["account_kit_sms_login_url"] = account_kit_sms_login_url(state, **kw)
    if hasattr(settings, "FACEBOOK_APP_ID") and settings.FACEBOOK_APP_ID:
        data["facebook_login_url"] = facebook_login_url(state, **kw)
    if hasattr(settings, "GOOGLE_APP_ID") and settings.GOOGLE_APP_ID:
        data["google_login_url"] = google_login_url(state, **kw)
    return data



from datetime import timedelta
from django.contrib.auth.models import User
from django.db import transaction
from django.utils.timezone import now
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.authtoken.models import Token
from .helpers import account_kit_me, account_kit_get_access_token
from .helpers import facebook_get_access_token, facebook_me
from .models import AccountKitUser, AccountKitAccessToken
from .models import FacebookUser, FacebookAccessToken
from .helpers import google_me, google_get_access_token
from .models import GoogleUser, GoogleAccessToken


class AuthResultSerializer(serializers.Serializer):
    def to_representation(self, instance: User):
        assert isinstance(instance, User)
        token = Token.objects.get_or_create(user=instance)[0]
        return {
            "token": token.key,
            "user": {"id": instance.id, "username": instance.username, "email": instance.email},
        }


class AccountKitAuthTokenSerializer(serializers.Serializer):
    code = serializers.CharField(label=_("Code"), required=False)
    access_token = serializers.CharField(label=_("Access Token"), required=False)

    def validate(self, attrs):
        code = attrs.get("code", "")
        access_token = attrs.get("access_token", "")
        expire_time = None

        if not access_token:
            if not code:
                raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (NO_CODE)", code="authorization")
            request_time = now()
            res = account_kit_get_access_token(code)
            if res.status_code != 200:
                raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (INVALID_OR_USED_CODE)", code="authorization")
            data = res.json()
            access_token = data["access_token"]
            expire_time = request_time + timedelta(seconds=data["token_refresh_interval_sec"])

        res = account_kit_me(access_token)
        if res.status_code != 200:
            raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (ME)", code="authorization")
        me = res.json()
        ak_user_id = me["id"]

        with transaction.atomic():
            fields = ["me"]
            ak_user, created = AccountKitUser.objects.get_or_create(ext_user_id=ak_user_id)
            if created:
                user = User(username="ak-{}".format(ak_user_id))
                user.save()
                ak_user.user = user
                fields.append("user")
            ak_user.me = me
            ak_user.save(update_fields=fields)
            ak_token, created = AccountKitAccessToken.objects.get_or_create(ext_user=ak_user, access_token=access_token, defaults={"expire_time": expire_time})
            if ak_token.expire_time != expire_time:
                ak_token.expire_time = expire_time
                ak_token.save(update_fields=["expire_time"])

        attrs["user"] = ak_user.user
        return attrs


class FacebookAuthTokenSerializer(serializers.Serializer):
    code = serializers.CharField(label=_("Code"), required=False)
    access_token = serializers.CharField(label=_("Access Token"), required=False)

    def validate(self, attrs):
        code = attrs.get("code", "")
        access_token = attrs.get("access_token", "")
        expire_time = None

        if not access_token:
            if not code:
                raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (NO_CODE)", code="authorization")
            request_time = now()
            res = facebook_get_access_token(code)
            if res.status_code != 200:
                raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (INVALID_OR_USED_CODE)", code="authorization")
            data = res.json()
            access_token = data["access_token"]
            expire_time = request_time + timedelta(seconds=data["expires_in"])

        res = facebook_me(access_token)
        if res.status_code != 200:
            raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (ME)", code="authorization")
        me = res.json()
        fb_user_id = me["id"]

        with transaction.atomic():
            fields = ["me"]
            fb_user, created = FacebookUser.objects.get_or_create(ext_user_id=fb_user_id)
            if created:
                user = User(username="fb-{}".format(fb_user_id))
                user.save()
                fb_user.user = user
                fields.append("user")
            fb_user.me = me
            fb_user.save(update_fields=fields)
            fb_token, created = FacebookAccessToken.objects.get_or_create(ext_user=fb_user, access_token=access_token, defaults={"expire_time": expire_time})
            if fb_token.expire_time != expire_time:
                fb_token.expire_time = expire_time
                fb_token.save(update_fields=["expire_time"])

        attrs["user"] = fb_user.user
        return attrs


class GoogleAuthTokenSerializer(serializers.Serializer):
    code = serializers.CharField(label=_("Code"), required=False)
    access_token = serializers.CharField(label=_("Access Token"), required=False)

    def validate(self, attrs):
        code = attrs.get("code", "")
        access_token = attrs.get("access_token", "")
        expire_time = None

        if not access_token:
            if not code:
                raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (NO_CODE)", code="authorization")
            request_time = now()
            res = google_get_access_token(code)
            if res.status_code != 200:
                raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (INVALID_OR_USED_CODE)", code="authorization")
            data = res.json()
            access_token = data["access_token"]
            expire_time = request_time + timedelta(seconds=data["expires_in"])

        res = google_me(access_token)
        if res.status_code != 200:
            raise serializers.ValidationError(_("Unable to log in with provided credentials.") + " (ME)", code="authorization")
        me = res.json()
        gg_user_id = me["id"]

        with transaction.atomic():
            fields = ["me"]
            gg_user, created = GoogleUser.objects.get_or_create(ext_user_id=gg_user_id)
            if created:
                user = User(username="gg-{}".format(gg_user_id))
                user.save()
                gg_user.user = user
                fields.append("user")
            gg_user.me = me
            gg_user.save(update_fields=fields)
            gg_token, created = GoogleAccessToken.objects.get_or_create(ext_user=gg_user, access_token=access_token, defaults={"expire_time": expire_time})
            if gg_token.expire_time != expire_time:
                gg_token.expire_time = expire_time
                gg_token.save(update_fields=["expire_time"])

        attrs["user"] = gg_user.user
        return attrs



from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.db.models import JSONField  # type: ignore
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _


class OAuthUser(models.Model):
    created = models.DateTimeField(_("created"), default=now, blank=True, db_index=True, editable=False)
    user = models.ForeignKey(User, on_delete=models.CASCADE, null=True, default=None, blank=True)
    ext_user_id = models.CharField(db_index=True, max_length=32)
    me = JSONField(blank=True, default=dict)

    class Meta:
        abstract = True


class OAuthAccessToken(models.Model):
    access_token = models.CharField(max_length=255)
    expire_time = models.DateTimeField(db_index=True)

    class Meta:
        abstract = True


class AccountKitUser(OAuthUser):
    @property
    def phone(self) -> str:
        return self.me.get("phone", {}).get("number", "") if self.me else ""


class AccountKitAccessToken(OAuthAccessToken):
    ext_user = models.ForeignKey(AccountKitUser, on_delete=models.CASCADE, related_name="+")


class FacebookUser(OAuthUser):
    @property
    def picture_url(self) -> str:
        return settings.FACEBOOK_API_URL + "/{}/picture?type=square".format(self.ext_user_id)

    @property
    def name(self) -> str:
        return self.me.get("name") if self.me else ""

    @property
    def first_name(self) -> str:
        name = self.name
        if name:
            parts = name.split(" ")
            return " ".join(parts[:-1])
        return ""

    @property
    def last_name(self) -> str:
        name = self.name
        if name:
            parts = name.split(" ")
            return " ".join(parts[-1:])
        return ""

    @property
    def phone(self) -> str:
        return self.me.get("phone", {}).get("number", "") if self.me else ""

    @property
    def email(self) -> str:
        return self.me.get("email", "") if self.me else ""


class FacebookAccessToken(OAuthAccessToken):
    ext_user = models.ForeignKey(FacebookUser, on_delete=models.CASCADE, related_name="+")


class GoogleUser(OAuthUser):
    @property
    def email(self) -> str:
        return self.me.get("email", "") if self.me else ""

    @property
    def verified_email(self) -> bool:
        return self.me.get("verified_email", False) if self.me else False

    @property
    def locale(self) -> str:
        return self.me.get("locale", "") if self.me else ""

    @property
    def picture_url(self) -> str:
        return self.me.get("picture", "") if self.me else ""

    @property
    def name(self) -> str:
        return self.me.get("name", "") if self.me else ""

    @property
    def first_name(self) -> str:
        return self.me.get("given_name", "") if self.me else ""

    @property
    def last_name(self) -> str:
        return self.me.get("family_name", "") if self.me else ""


class GoogleAccessToken(OAuthAccessToken):
    ext_user = models.ForeignKey(GoogleUser, on_delete=models.CASCADE, related_name="+")
